/*
 * SEISANARCH
 * Author: R. Luis
 * Purpose: c functions for accessing mini-seed archives containing dayfiles
 * Last Update: 2010.11.15
 * 
 * Main Functions:
 * getarchdata_ - Populates an array with samples for a given request
 * getarchinfo_ - Computes the number of samples that may be extracted by getarchdata_
 *
 * Notes:
 * These functions are prepared to work only with archives of with the following structures:
 * Buffer of Uniform Data (BUD) - <BUDdir>/<NET>/<STA>/STA.NET.LOC.CHAN.YEAR.DAY
 * SeisComP Data Structure (SDS) - <SDSdir>/<YEAR>/<NET>/<STA>/<CHAN.TYPE>/NET.STA.LOC.CHAN.TYPE.YEAR.DAY
 *
 * History of Changes:
 * 2010.11.15
 *   Removed BUD-specific function to compute filename and added generic naming funcion
 *   This allows usage of SDS and BUD. Eventually other structures may be added.
 * 2010.10.14
 *   First version
 *   Reads only from BUD structures, requires adaptation for SDS as well.
 *
 * List of TODOs
 * - Include bissection algorithm for searching data, current algorithm is slow at end of file
 * - Allow more generic search with more than two files for eventually adding SEISAN databases
 * - Perform data search without loading data. Might be even faster.
 *
*/

#include <libmseed.h>
#include <stdio.h>
#include <string.h>

/* Declarations */
char *trim_string(char *original, int nchar);
char *getArchFilename(hptime_t time, char *filename, char *af, char *sta, char *chan, char *net, char *loc, int *archiveType);
//char *getSDSFilename(hptime_t time, char *filename, char *af, char *sta, char *chan, char *net, char *loc);
void getMSTG(hptime_t starttime, hptime_t endtime, char *file1, char *file2, MSTraceGroup *mstg, int dataflag);

/*
getarchinfo
Return base data from a BUD archive
If data for the requested station and time interval is unavailable, returns zero number of samples. 
*/
void getarchinfo_(archiveFolder, archiveFolderLen, archiveType, stat, statlen, chan, chanlen, net, netlen, loc, loclen, year, month, day, hour, min, sec, t_interval,
	nsamp, samprate, t_error, nyear, nmonth, nday, nhour, nmin, nsec)
//Input arguments
char *archiveFolder;	//Name of the base folder of the archive
int *archiveFolderLen;	//Length of the archive folder name
int *archiveType;	//Integer indicating the type of archive to access
			//   0 - BUD Archive
			//   1 - SeisComP Archive
char *stat; 		//station name
int *statlen; 		//station name length
char *chan; 		//channel name
int *chanlen;		//channel name length
char *net; 		//network name
int *netlen;		//network name length
char *loc;		//location name
int *loclen;		//location name length
int *year;		//start time year 
int *month;		//start time month
int *day;		//start time day of month
int *hour;		//start time hour
int *min;		//start time minute
float *sec;		//start time second
int *t_interval;	//Requested time interval
//Output arguments
int *nsamp;		//Number of available samples for the requested interval
float *samprate;	//Sampling rate
int *t_error;		//Timing error
int *nyear;
int *nmonth;
int *nday;
int *nhour;
int *nmin;
float *nsec;
{
	//Set default output parameters in case no data is found
	*nsamp = 0;
	*samprate = 0.0f;
	*t_error = 0;


	//Calculate start and end times of the request
	int doy_start; 		//Start day of year
	int doy_end;		//End day of year
	hptime_t starttime;	//Start time in format from libmseed
	hptime_t endtime;	//Same as starttime
	BTime bstarttime;	//Start time in BTime format (SEED specification)
	BTime bendtime; 	//Save as bstarttime

	ms_md2doy(*year, *month, *day, &doy_start); //Compute doy for starttime
	
	//Compute start and end times
	starttime = ms_time2hptime(*year, doy_start, *hour, *min, 0, 0); 
	endtime = starttime + (hptime_t)(*t_interval) * (hptime_t)(1000000);
	
	//Compute btimes
	ms_hptime2btime(starttime, &bstarttime); 
	ms_hptime2btime(endtime, &bendtime);
	
	//Determine file names
	char fstart[120]; //Start file name
	char fend[120];  //End file name
	//Prepare char arrays with input data
	char *tarch = trim_string(archiveFolder, *archiveFolderLen);    //Base folder
	char *tnet = trim_string(net, *netlen);				//Network code
	char *tstat = trim_string(stat, *statlen);			//Station code
	char *tloc = trim_string(loc, *loclen);				//Location code
	char *tchan = trim_string(chan, *chanlen);			//Channel code
	
	//Get filename, according to archive type
	getArchFilename(starttime, fstart, tarch, tstat, tchan, tnet, tloc, archiveType);
	getArchFilename(starttime, fstart, tarch, tstat, tchan, tnet, tloc, archiveType);
	//printf("Filenames:\n <%s>\n <%s>\n", fstart, fend); //For debugging
	
	//Initiate MSTraceGroup
	MSTraceGroup *mstg = 0;
  	mstg = mst_initgroup (mstg);
	
	//Read sequentially all record headers without including data
	getMSTG(starttime, endtime, fstart, fend, mstg, 0);
	
	//mst_printtracelist (mstg, 0, 1, 1); //For debugging
           	
        //Check for more than one read trace//Compute btimes
        if (mstg->numtraces<1)
        	return;

	//Calculate samples to skip at the start and correct starttime
	int sampleSkip = 0;
	MSTrace *mst = mstg->traces;	
	
	//char strx[120]; printf("Old starttime: %s\n", ms_hptime2mdtimestr(starttime, strx, 1)); //For debugging
	if (starttime > mst->starttime) 
	{
		sampleSkip = (int)(((starttime - mst->starttime) * mst->samprate)/1000000);
		starttime = mst->starttime + ( sampleSkip * 1000000 ) / mst->samprate;
	}
	else
	{
		starttime = mst->starttime;
	}
	//char stry[120]; printf("New starttime: %s\n", ms_hptime2mdtimestr(starttime, strx, 1)); //For debugging
	
	
	//Isolate samplerate -> TODO USE NOMINAL SAMPLING RATE!!!
	*samprate = (float)mst->samprate;
		
	//Count number of samples including gaps
	*nsamp = -sampleSkip;
	hptime_t lastEndtime = mst->starttime;
	while (mst != 0)
	{	
		//Include gaps
		*nsamp += (mst->starttime - lastEndtime) * mst->samprate / 1000000;
		int gapsamp = (mst->starttime - lastEndtime) * mst->samprate / 1000000;
		//Intermediate or start trace
		if (endtime >= mst->endtime)
		{
			*nsamp += mst->samplecnt;
			lastEndtime = mst->endtime + (1000000/mst->samprate);
		}
		else
		{
			*nsamp += (endtime - mst->starttime) * mst->samprate / 1000000;
			break;
		}
		mst = mst->next;
	}
	
	//printf("Output samples: %d\n", *nsamp);
	//printf("Sampling rate: %d\n", *samprate);
	
	
	//Produce new starttime output args
	char ts[27];
	ms_hptime2isotimestr(starttime, ts, 1);
	sscanf(ts, "%4d-%2d-%2dT%2d:%2d:%9f",
		nyear, nmonth, nday, nhour, nmin, nsec);

        // Cleanup memory and close file
        mst_freegroup (&mstg);
}













/*
get_BUDData
Return sample data from a BUD archive
If data for the requested station and time interval is unavailable, returns samples as zero. 
*/
void getarchdata_(archiveFolder, archiveFolderLen, archiveType, stat, statlen, chan, chanlen, net, netlen, loc, loclen, year, month, day, hour, min, sec, t_interval,
	samples)
//Input arguments
char *archiveFolder;	//Name of the base folder of the BUD archive
int *archiveFolderLen;	//Length of the archive folder name
int *archiveType;       //Archive type
char *stat; 		//station name
int *statlen; 		//station name length
char *chan; 		//channel name
int *chanlen;		//channel name length
char *net; 		//network name
int *netlen;		//network name length
char *loc;		//location name
int *loclen;		//location name length
int *year;		//start time year 
int *month;		//start time month
int *day;		//start time day of month
int *hour;		//start time hour
int *min;		//start time minute
float *sec;		//start time second
int *t_interval;	//Requested time interval
//Output arguments
int *samples;		//Array with the requested samples
{
	//printf("\n     ##### STARTING GETBUDDATA #####\n");
	//Calculate start and end times
	int doy_start;
	int doy_end;
	hptime_t starttime;
	hptime_t endtime;
	BTime bstarttime;
	BTime bendtime; 

	ms_md2doy(*year, *month, *day, &doy_start); //Compute doy for starttime
	starttime = ms_time2hptime(*year, doy_start, *hour, *min, 0, 0); //Compute start and end times
	endtime = starttime + (hptime_t)(*t_interval) * (hptime_t)(1000000);
	
	ms_hptime2btime(starttime, &bstarttime); //Compute btimes
	ms_hptime2btime(endtime, &bendtime);
	
	//Determine file names
	char fstart[120]; //Start file
	char fend[120];  //End file
	char *tarch = trim_string(archiveFolder, *archiveFolderLen);
	char *tnet = trim_string(net, *netlen);
	char *tstat = trim_string(stat, *statlen);
	char *tloc = trim_string(loc, *loclen);
	char *tchan = trim_string(chan, *chanlen);	
	
	//Get filename, according to archive type
	getArchFilename(starttime, fstart, tarch, tstat, tchan, tnet, tloc, archiveType);
	getArchFilename(starttime, fstart, tarch, tstat, tchan, tnet, tloc, archiveType);
	//printf("Filenames:\n <%s>\n <%s>\n", fstart, fend); //For debugging
	
	//Init MSTraceGroup
	MSTraceGroup *mstg = 0;
  	mstg = mst_initgroup (mstg);
	
	//Read sequentially all record headers without including data
	getMSTG(starttime, endtime, fstart, fend, mstg, 1);
	//mst_printtracelist (mstg, 0, 1, 1); //For debugging
           	
        //Check for more than one read trace
        if (mstg->numtraces<1)
        	return;        
        
        //Initialize first trace
        MSTrace *mst = mstg->traces;
        	
	//Calculate samples to skip at the start and correct starttime
	//char strx[120]; printf("Old starttime: %s\n", ms_hptime2mdtimestr(starttime, strx, 1)); //For debugging
	int sampleSkip = 0;
	if (starttime > mst->starttime) 
	{
		sampleSkip = (int)(((starttime - mst->starttime) * mst->samprate)/1000000);
		starttime = mst->starttime + ( sampleSkip * 1000000 ) / mst->samprate;
	}
	else
	{
		starttime = mst->starttime;
	}
			
	//Count number of samples including gaps
	int *curSample = samples;
	hptime_t lastEndtime = mst->starttime;
	int scount; //Counter for sample fill
	int posStart;
	int posEnd;
	int gap;
	while (mst != 0)
	{
		//Check if zeros need to be introduced
		gap = ((mst->starttime - lastEndtime)*mst->samprate)/1000000; //Gap in samples
		
		//Fill in gap
		for (scount = 0; scount<gap; scount++)
			*(curSample++) = 0;
		
		//Find out which samples should be extracted from this mst
		posStart = (starttime <= mst->starttime) ? 0 : (((starttime - mst->starttime)*mst->samprate)/1000000);
		posEnd = (endtime>=mst->endtime)?mst->numsamples:(((endtime - mst->starttime)*mst->samprate)/1000000);
		
		//Copy samples
		for (scount = posStart; scount<posEnd; scount++)
			*(curSample++) = *((int*)mst->datasamples + scount);
		
		//Update last endtime
		lastEndtime = mst->endtime + 1000000/mst->samprate;
		
		//Go for next mst
		mst = mst->next;
	}

	//printf("Output samples: %d\n", curSample - samples);
	
	
	//Print samples in ASCII format for testing
	/*
	FILE * pFile;
	int n;
   	char name [100];
   	pFile = fopen ("samples.txt","w");
   	for (n=0 ; n<(curSample - samples) ; n++)
   	{
     		fprintf (pFile, "%d %d\n",n, samples[n]);
   	}
   	fclose (pFile);
	*/
	
	// Cleanup memory and close file
        mst_freegroup (&mstg);
}





















/*
Function to trim a char string and limit the number of characters
*/
char *trim_string(char *original, int nchar)
{
	int i;
	for (i=0; i<nchar; i++)
		if (original[i] == ' ')
			break;
	char *trimmed = malloc(i);
	trimmed[i] = '\0';
	strncpy(trimmed, original, i);
	return trimmed;
}

/*
Function to compute Archive filenames
*/
char *getArchFilename(hptime_t time, char *filename, char *af, char *stat, char *chan, char *net, char *loc, int *archiveType)
{
	BTime btime;
	ms_hptime2btime(time, &btime); //Compute btime
	
	//Determine file name
	if ( *archiveType == 0 ) 
	{
		//BUD Archive
		sprintf(filename, "%s/%s/%s/%s.%s.%s.%s.%d.%d", 
			af, net, stat, stat, net, loc, chan, btime.year, btime.day);
	} 
	else if ( *archiveType == 1 )
	{
		//SDS Archive
		sprintf(filename, "%s/%d/%s/%s/%s.D/%s.%s.%s.%s.D.%d.%d", 
			af, btime.year, net, stat, chan, net, stat, loc, chan, btime.year, btime.day);
	}
	else 
	{
		//Unknown archive type, leave
		return NULL;
	}	
	return filename;
}

/*
Function to read a set of records in one or two files and return mstg
*/
void getMSTG(hptime_t starttime, hptime_t endtime, char *file1, char *file2, MSTraceGroup *mstg, int dataflag)
{
	//Read headers
	MSRecord *msr = NULL;
	int retcode;
	hptime_t recstart;
	hptime_t recend;
	int fcount;
	int nFiles = 2; //Number of files to check
	char *curFilename = file1;
	
	//Compare file names
	if (strcmp(file1,file2) == 0)
		nFiles = 1; //Identical filenames, check only one
	
	
	for (fcount = 0; fcount<nFiles; fcount++)
	{
		//printf("Reading file: <%s>\n", curFilename);
		while ((retcode = ms_readmsr(&msr, curFilename, 0, NULL, NULL, 1, dataflag, 0)) == MS_NOERROR)
		{
			//msr_print(msr,0); //For debugging
			
			//Check if trace is relevant for request
			recstart = msr->starttime;
			recend = msr->starttime + (hptime_t)(msr->samplecnt * 1000000 / msr->samprate);

			//Check if there is any need to continue
			if (recstart>endtime)
				break;

			//Add record to trace group
			if ((recstart<=endtime) && (recend >= starttime))
			{
				mst_addmsrtogroup(mstg, msr, 0, 0, 0);
				
				//msr_print(msr,0); //For debugging
				
				//Check time correction only for the first record
				/*				
				if ((get_time_correction) && ((msr->fsdh->io_flags&128)>>7))
				{
					*t_error = msr->fsdh->time_correct;
					get_time_correction = 0;
				}
				*/
			}
		}
		ms_readmsr (&msr, NULL, 0,NULL, NULL, 0, 0, 0); //Reset record and free file
		curFilename = file2; //Set second file name
		
		//Check if it is necessary to go for another file
		if (endtime < recend)
			break;
	}
}




